/*******************************************************************************
 * 
 * jKiwi -- A virtual makeover and hairstyler application
 * Copyright (c) 2007-2008 Dan Mihai Ile, Maria Joao Roque Barbado Leal
 * 
 * This file is part of jKiwi.
 * 
 * jKiwi is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 * 
 * jKiwi is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 *******************************************************************************/

package core;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.Graphics2D;
import java.awt.Rectangle;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;
import java.awt.geom.AffineTransform;
import java.awt.geom.Rectangle2D;
import java.awt.image.AffineTransformOp;
import java.awt.image.BufferedImage;
import java.io.File;
import java.io.FilenameFilter;
import java.util.ResourceBundle;


import org.eclipse.swt.SWT;
import org.eclipse.swt.SWTException;
import org.eclipse.swt.graphics.GC;
import org.eclipse.swt.graphics.ImageData;
import org.eclipse.swt.graphics.ImageLoader;
import org.eclipse.swt.graphics.PaletteData;
import org.eclipse.swt.graphics.Point;
import org.eclipse.swt.graphics.RGB;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Control;
import org.eclipse.swt.widgets.Event;
import org.eclipse.swt.widgets.FileDialog;
import org.eclipse.swt.widgets.MessageBox;
import org.eclipse.swt.widgets.Shell;
import org.eclipse.swt.widgets.Menu;
import org.eclipse.swt.widgets.MenuItem;
import org.eclipse.swt.graphics.Image;
import org.eclipse.swt.layout.FillLayout;
import org.eclipse.swt.widgets.Display;

import core.MainWindow;
import dialogs.BeforeAndAfter;
import dialogs.FileIOProgress;
import dialogs.FitImage;
import dialogs.FullScreenPreview;
import tabs.TabsContainer;
import utils.ImageUtils;
import utils.StockIcon;

/**
 * Instances of this class represents a project in use by the main window.
 * Every time a image file is opened by the main window a instance of this
 * class is created to deal with all the aspects of the image.
 */
public class ProjectWindow {
	
	private class NoHiddenFilesFilter implements FilenameFilter {

		public boolean accept(File dir, String name) {
			if (name.startsWith(".") || new File(dir + File.separator + name).isHidden()) {
				return false;
			} else {
				return true;
			}
		}
	}

	/**
	 * The width and height that the project
	 * will use to work with the image 
	 */
	private final int RESIZED_IMAGE_WIDTH = 384;
	private final int RESIZED_IMAGE_HEIGHT = 512;

	/**
	 * the guessed position of the eyes in the image
	 */
	private final Rectangle2D.Double eyesPosition = new Rectangle2D.Double(0.26, 0.4, 0.47, 0.15);

	/**
	 * the guessed position of the lips in the image
	 */
	private final Rectangle2D.Double lipsPosition = new Rectangle2D.Double(0.26, 0.6, 0.47, 0.15);

	/**
	 * The maximum allowed history steps for doing undo
	 */
	public final int MAX_UNDO_HISTORY = 40;

	/**
	 * The tabs used by this instance of ProjectWindow
	 */
	private TabsContainer tabs = null;

	/**
	 * The shell used to show this project's window
	 */
	private Shell sShell = null;

	/**
	 * The parent MainWindow
	 */
	private MainWindow parent = null;

	/**
	 * messages that are to be used by the entire GUI
	 */
	private ResourceBundle messages;

	//all menu variables here
	private Menu menuBar = null;
	private Menu submenuProject = null;
	private Menu submenuEdit = null;
	private Menu submenuView = null;
	private Menu submenuAdjustments;
	private MenuItem pushUndo = null;
	private MenuItem pushRedo = null;
	private MenuItem pushClear = null;
	private MenuItem pushFullScreen;
	private MenuItem showHairMenu;
	private MenuItem pushEyesTracer;
	private MenuItem pushLensTracer;
	private MenuItem pushLipsTracer;

	/**
	 * The class in charge for displaying the image
	 */
	private BridgeAWT_SWT imageArea = null;

	/**
	 * Image temporarily used to store the (possibly large) original image data read from the file
	 */
	private BufferedImage tempOriginalImage;

	/**
	 * The path to the image in use by this project
	 */
	private String selectedImagePath;

	/**
	 * The resulting image that will be used by this project
	 */
	private BufferedImage resizedFitImage;

	/**
	 * The full screen window that will be used by the project
	 */
	private FullScreenPreview fullScreenPreview = null;

	/**
	 * When the project is first saved here we keep a reference to the saved file
	 */
	private String selectedFileForSave = null;

	/**
	 * The high quality eyes image area
	 */
	public SubImage eyesSubImage;

	/**
	 * The high quality lips image area
	 */
	public SubImage lipsSubImage;

	/**
	 * Creates a new instance of this class given it's parent,
	 * and the name of the file to be used by this project which
	 * if null then this class asks for a file using a FileDialog
	 * @param parent the main window parent
	 * @param fileName the file that should be opened by this instance or null
	 */
	public ProjectWindow (MainWindow parent, String fileName) {
		this.parent = parent;
		messages = Loader.getMessages();
		readImage(fileName);
	}

	/**
	 * Read a given file from the file system and creates the
	 * {@link FitImage} dialog so that the user adjust it.
	 * @param fileName the image file path and name
	 * @return a rectangle with the clipping coordinates and the angle to rotate the image
	 */
	private Object[] fitImage(String fileName) {

		//TODO what if file does not exist?

		//read the file from the file system
		FileIOProgress fp = new FileIOProgress(parent.getSShell());
		BufferedImage readedImage = fp.openImage(fileName);
		
		if (readedImage == null) { //no file was read
			return null;
		} else { //file was read, proceed to fit the image

			//create a white border around the image so that when it will not get clipped in corners when rotated
			int size = (int) Math.sqrt(Math.pow(readedImage.getWidth(), 2) + Math.pow(readedImage.getHeight(), 2));
			tempOriginalImage = new BufferedImage(size, size, BufferedImage.TYPE_INT_RGB);
			Graphics2D g2d = tempOriginalImage.createGraphics();
			g2d.setColor(Color.WHITE);
			g2d.fillRect(0, 0, tempOriginalImage.getWidth(), tempOriginalImage.getHeight());
			g2d.drawImage(readedImage, null, (tempOriginalImage.getWidth() / 2) - (readedImage.getWidth() / 2) , (tempOriginalImage.getHeight() / 2) - (readedImage.getHeight() / 2));
			g2d.dispose();
			readedImage = null;
			
			FitImage fi = new FitImage(parent.getSShell(), tempOriginalImage);
			final Object[] fitImageData = fi.open();
			if (fitImageData == null) { //cancel was pressed during fitting the image
				//TODO memory leaks?
				return null;
			} else {
				Rectangle cuttedImageArea = (Rectangle) fitImageData[0];
				Dimension imageSize = (Dimension) fitImageData[1];
				Double imageAngle = (Double) fitImageData[2];
				int xOffset = (int)Math.round((tempOriginalImage.getWidth() * cuttedImageArea.x) / (double)imageSize.width);
				int yOffset = (int)Math.round((tempOriginalImage.getHeight() * cuttedImageArea.y) / (double)imageSize.height);
				int subImageWidth = (int)Math.round((tempOriginalImage.getWidth() * cuttedImageArea.width) / (double)imageSize.width);
				int subImageHeight = (int)Math.round((tempOriginalImage.getHeight() * cuttedImageArea.height) / (double)imageSize.height);
				return new Object[] { new Rectangle(xOffset, yOffset, subImageWidth, subImageHeight), imageAngle };
			}
		}
	}

	/**
	 * Read the given file from the file system or
	 * ask the user for a file if null is given
	 * @param file the file to be read from the file system or null
	 */
	private void readImage(String file) {

		if (file == null) {
			//create a file dialog to ask for the image
			FileDialog fd = new FileDialog(parent.getSShell(), SWT.OPEN);
			String openPath = null;
			try {
				openPath = new File(System.getProperty("user.home")).listFiles(new NoHiddenFilesFilter())[0].getAbsolutePath();				
			} catch (Exception e) {
				openPath = System.getProperty("user.home");
			}
			fd.setFileName(openPath);
			fd.setText(messages.getString("PROJECT_WINDOW_REQUEST_IMAGE-DIALOG_TITLE"));
			fd.setFilterNames(new String[] { FileDialogFilter.AllImagesName(), FileDialogFilter.jpegName(), FileDialogFilter.pngName() });
			fd.setFilterExtensions(new String[] { FileDialogFilter.AllImagesExtension(), FileDialogFilter.jpegExtension(), FileDialogFilter.pngExtension() });		
			selectedImagePath = fd.open();
		} else {
			selectedImagePath = file;
		}

		//call for fit image before opening the project window
		Object[] fitImageData = fitImage(selectedImagePath);
		if (fitImageData != null) { //fitImage terminated successfully

			parent.getSShell().setCursor(Display.getCurrent().getSystemCursor(SWT.CURSOR_WAIT));

			//create the tabs and the imageArea so the fitted image is drawn for the first time
			tabs = new TabsContainer(this,SWT.NONE, RESIZED_IMAGE_WIDTH, RESIZED_IMAGE_HEIGHT);
			createSShell();
			tabs.createTabFolder();

			Double imageAngle = (Double) fitImageData[1];
			Rectangle subImageOffset = (Rectangle) fitImageData[0];

			BufferedImage tmp = rotateAndClip(imageAngle, subImageOffset);
			resizedFitImage = getResizeImage(tmp, RESIZED_IMAGE_WIDTH, RESIZED_IMAGE_HEIGHT, java.awt.Image.SCALE_SMOOTH);
			tempOriginalImage = null;
			eyesSubImage = new SubImage(resizedFitImage, eyesPosition, 3, tmp);
			lipsSubImage = new SubImage(resizedFitImage, lipsPosition, 3, tmp);
			tmp = null;

			tabs.setOriginalImage(resizedFitImage);
			System.gc();

			//hide all other project's tabs on main window (or any other Control that is visible on MainContainer)
			for (Control c : parent.getMainContainer().getChildren()) {
				c.setVisible(false);
			}

			//open the project's window
			sShell.open();

			//TODO keep aspect ratio when creating the icon
			sShell.setImage(new Image(sShell.getDisplay(),new ImageLoader().load(selectedImagePath)[0].scaledTo(128, 128)));
			parent.getSShell().setCursor(Display.getCurrent().getSystemCursor(SWT.DEFAULT));
		}
	}

	/**
	 * Rotate the given image and then clip it using the given rectangle
	 * @param imageAngle the angle which the image will be rotated
	 * @param subImageOffset the sub image coordinates used for clipping
	 * @return the resulting clipped and rotated image
	 */
	private BufferedImage rotateAndClip(Double imageAngle, Rectangle subImageOffset) {

		//rotate the original image if necessary
		if (imageAngle != 0.0) {
			AffineTransform tx = new AffineTransform();
			tx.rotate(Math.toRadians(imageAngle), tempOriginalImage.getWidth() / 2.0, tempOriginalImage.getHeight() / 2.0);
			AffineTransformOp op = new AffineTransformOp(tx, AffineTransformOp.TYPE_BICUBIC);
			tempOriginalImage = op.filter(tempOriginalImage, null);
		}
		
		//check if required portion of image is within the image bounds
		if (subImageOffset.x < 0 || subImageOffset.y < 0 ||
				subImageOffset.x + subImageOffset.width > tempOriginalImage.getWidth() ||
				subImageOffset.y + subImageOffset.height > tempOriginalImage.getHeight()) {
			BufferedImage res = new BufferedImage(subImageOffset.width, subImageOffset.height, BufferedImage.TYPE_INT_RGB);
			Graphics2D g2d = res.createGraphics();
			g2d.setColor(Color.WHITE);
			g2d.fillRect(0, 0, res.getWidth(), res.getHeight());
			g2d.drawImage(tempOriginalImage, -subImageOffset.x, -subImageOffset.y, null);
			g2d.dispose();
			g2d = null;
			return res;
		} else {
			return tempOriginalImage.getSubimage(subImageOffset.x, subImageOffset.y, subImageOffset.width , subImageOffset.height);
		}
	}

	/**
	 * Returns a new resized image using the
	 * {@value AffineTransformOp#TYPE_NEAREST_NEIGHBOR}
	 * given a width and height factor for resizing
	 * @param image the image to be resized
	 * @param widthFactor the desired width factor
	 * @param heightFactor the desired height factor
	 * @return a new resized image
	 */
	private BufferedImage getResizedImage_Nearest(BufferedImage image, double widthFactor, double heightFactor) {
		AffineTransform tx = new AffineTransform(); //use affine transform class to do the resizing
		tx.scale(widthFactor, heightFactor);
		AffineTransformOp op = new AffineTransformOp(tx, AffineTransformOp.TYPE_NEAREST_NEIGHBOR);
		return op.filter(image, null);
	}

	/**
	 * Returns a new resized image using the
	 * {@value java.awt.Image#SCALE_SMOOTH}
	 * given a width and height factor for resizing
	 * @param image the image to be resized
	 * @param widthFactor the desired width factor
	 * @param heightFactor the desired height factor
	 * @return a new resized image
	 */
	private BufferedImage getResizedImage_Smooth(BufferedImage image, double widthFactor, double heightFactor) {
		java.awt.Image img1 = image.getScaledInstance((int)Math.round(image.getWidth()*widthFactor), (int)Math.round(image.getHeight()*heightFactor), java.awt.Image.SCALE_SMOOTH);
		BufferedImage bFinal = new BufferedImage((int)Math.round(image.getWidth()*widthFactor), (int)Math.round(image.getHeight()*heightFactor), BufferedImage.TYPE_INT_ARGB);
		Graphics2D g2d = bFinal.createGraphics();
		g2d.drawImage(img1, 0, 0, null);
		g2d.dispose();
		img1 = null;
		return bFinal;
	}

	/**
	 * Resize the given image to the desired width and height,
	 * using the resize quality which could be one of the
	 * following algorithms.
	 * @see AffineTransformOp#TYPE_NEAREST_NEIGHBOR
	 * @see java.awt.Image#SCALE_SMOOTH
	 * 
	 * @param image the image to be resized
	 * @param width the desired image width
	 * @param height the desired image height
	 * @param resizeQuality the resizing quality to be used on resizing
	 * @return a new resized image
	 */
	private BufferedImage getResizeImage(BufferedImage image, int width, int height, int resizeQuality) {

		double ratio = ImageUtils.getResizingRatio(image, width, height);

		BufferedImage resizedImage = null;
		switch (resizeQuality) {
		case AffineTransformOp.TYPE_NEAREST_NEIGHBOR:
			resizedImage =  getResizedImage_Nearest(image, ratio, ratio);
			break;
		case java.awt.Image.SCALE_SMOOTH:
			resizedImage =  getResizedImage_Smooth(image, ratio, ratio);
			break;

		default:
			break;
		}

//		//creates an empty image that is used to store the final result
//		BufferedImage result = new BufferedImage(width, height, BufferedImage.TYPE_INT_RGB);
//		Graphics2D g2d = result.createGraphics();
//		//TODO WARNING: definitely need a fix, fitImage also...
//		//TODO better fix?: sometimes the image has size = 382.98 x 512.0 instead of 384 x 512 so we stretch the image with one pixel
//		//g2d.drawImage(resizedImage, 0,0, width, height, null);
//
//		g2d.drawImage(resizedImage, 0,0, null);
//		g2d.dispose();
//		resizedImage = null;
//		System.gc();
		return resizedImage;
	}

	/**
	 * This method initializes sShell
	 */
	private void createSShell() {
		sShell = new Shell(SWT.DIALOG_TRIM);
		sShell.setText(messages.getString("PROJECT_WINDOW_TITLE"));
		imageArea = new BridgeAWT_SWT(this,SWT.NONE, tabs, RESIZED_IMAGE_WIDTH, RESIZED_IMAGE_HEIGHT);
		sShell.setLayout(new FillLayout());
		//if window is activated
		sShell.addShellListener(new org.eclipse.swt.events.ShellAdapter() {   
			public void shellActivated(org.eclipse.swt.events.ShellEvent e) {
				if (!parent.getSShell().isDisposed()) {
					//hide all other project's tabs on main window (or any other Control that is visible on MainContainer)
					for (Control c : parent.getMainContainer().getChildren()) {
						c.setVisible(false);
					}
					//TODO is there a better place to put this 2 lines of code!?
					tabs.pack();
					parent.getSShell().pack();
					tabs.setVisible(true);
				}
			}
			public void shellClosed(org.eclipse.swt.events.ShellEvent e) {
				sShell.getImage().dispose();
				tabs.dispose();
				//notify the main window that the project's tab was disposed
				parent.tabDisposed();
				//TODO free memory used by swing/awt drawing
				imageArea.dispose();
				sShell.dispose();
				eyesSubImage.dispose();
				lipsSubImage.dispose();

				tempOriginalImage = null;
				resizedFitImage = null;

				System.gc(); //TODO seems to make the difference
			}
		});

		sShell.setSize(new Point(RESIZED_IMAGE_WIDTH + 20, RESIZED_IMAGE_HEIGHT + 70));
		menuBar = new Menu(sShell, SWT.BAR);
		MenuItem pushProject = new MenuItem(menuBar, SWT.CASCADE);
		pushProject.setText(messages.getString("PROJECT_WINDOW_MENU_PROJECT"));
		MenuItem pushEdit = new MenuItem(menuBar, SWT.CASCADE);
		pushEdit.setText(messages.getString("PROJECT_WINDOW_MENU_EDIT"));
		MenuItem pushView = new MenuItem(menuBar, SWT.CASCADE);
		pushView.setText(messages.getString("PROJECT_WINDOW_MENU_VIEW"));
		submenuView = new Menu(pushView);
		pushView.setMenu(submenuView);
		MenuItem pushAdjustments = new MenuItem(menuBar, SWT.CASCADE);
		pushAdjustments.setText(messages.getString("PROJECT_WINDOW_MENU_ADJUSTMENTS"));
		submenuAdjustments = new Menu(pushAdjustments);
		pushAdjustments.setMenu(submenuAdjustments);
		MenuItem push8 = new MenuItem(submenuAdjustments, SWT.PUSH);
		push8.setText(messages.getString("PROJECT_WINDOW_MENU_ADJUSTMENTS_FITIMAGE"));
		push8.addSelectionListener(new org.eclipse.swt.events.SelectionAdapter() {
			public void widgetSelected(org.eclipse.swt.events.SelectionEvent e) {
				Object[] fitImageData = fitImage(selectedImagePath);
				if (fitImageData != null) { //fitImage terminated successfully

					//clear any previous changes
					tabs.clear();
					tabs.originalImageChanged(); //notify tabs that the original image was changed
					Double imageAngle = (Double) fitImageData[1];
					Rectangle subImageOffset = (Rectangle) fitImageData[0];

					BufferedImage tmp = rotateAndClip(imageAngle, subImageOffset);
					resizedFitImage = getResizeImage(tmp, RESIZED_IMAGE_WIDTH, RESIZED_IMAGE_HEIGHT, java.awt.Image.SCALE_SMOOTH);
					tempOriginalImage = null;
					eyesSubImage = new SubImage(resizedFitImage, eyesPosition, 3, tmp);
					lipsSubImage = new SubImage(resizedFitImage, lipsPosition, 3, tmp);
					tabs.setOriginalImage(resizedFitImage);
					imageArea.redrawDrawingContainer();
					tempOriginalImage = null;
					tmp = null;
					System.gc();
				}
			}
		});
		@SuppressWarnings("unused")
		MenuItem separator3 = new MenuItem(submenuAdjustments, SWT.SEPARATOR);
		pushEyesTracer = new MenuItem(submenuAdjustments, SWT.PUSH);
		pushEyesTracer.setEnabled(false);
		pushEyesTracer.setText(messages.getString("PROJECT_WINDOW_MENU_ADJUSTMENTS_EYESTRACER"));
		pushEyesTracer.addSelectionListener(new org.eclipse.swt.events.SelectionAdapter() {
			public void widgetSelected(org.eclipse.swt.events.SelectionEvent e) {
				tabs.setShadowPath();
			}
		});
		pushLensTracer = new MenuItem(submenuAdjustments, SWT.PUSH);
		pushLensTracer.setEnabled(false);
		pushLensTracer.setText(messages.getString("PROJECT_WINDOW_MENU_ADJUSTMENTS_LENSTRACER"));
		pushLensTracer.addSelectionListener(new org.eclipse.swt.events.SelectionAdapter() {
			public void widgetSelected(org.eclipse.swt.events.SelectionEvent e) {
				tabs.setLensPath();
			}
		});
		pushLipsTracer = new MenuItem(submenuAdjustments, SWT.PUSH);
		pushLipsTracer.setEnabled(false);
		pushLipsTracer.setText(messages.getString("PROJECT_WINDOW_MENU_ADJUSTMENTS_LIPSTRACER"));
		pushLipsTracer.addSelectionListener(new org.eclipse.swt.events.SelectionAdapter() {
			public void widgetSelected(org.eclipse.swt.events.SelectionEvent e) {
				tabs.setLipsPath();
			}
		});
		pushFullScreen = new MenuItem(submenuView, SWT.PUSH);
		pushFullScreen.setText(messages.getString("PROJECT_WINDOW_MENU_VIEW_FULL_SCREEN") + "\tF11");
		pushFullScreen.setAccelerator(SWT.F11);
		pushFullScreen.setImage(new Image(Display.getCurrent(), StockIcon.getIcon(StockIcon.ICON_FULLSCREEN, StockIcon.SIZE_MENU, true)));
		pushFullScreen.addSelectionListener(new org.eclipse.swt.events.SelectionAdapter() {
			public void widgetSelected(org.eclipse.swt.events.SelectionEvent e) {
				if (fullScreenPreview == null) {
					fullScreenPreview = new FullScreenPreview(sShell, imageArea.getCurrentImage());
				} else {
					if (fullScreenPreview.getSShell().isDisposed()) {
						fullScreenPreview = new FullScreenPreview(sShell, imageArea.getCurrentImage());
					}
				}
			}
		});
		MenuItem pushBeforeAfter = new MenuItem(submenuView, SWT.PUSH);
		pushBeforeAfter.setText(messages.getString("PROJECT_WINDOW_MENU_VIEW_BEFORE-AND-AFTER"));
		pushBeforeAfter.addSelectionListener(new org.eclipse.swt.events.SelectionAdapter() {
			public void widgetSelected(org.eclipse.swt.events.SelectionEvent e) {
				new BeforeAndAfter(sShell, ImageUtils.convertToSWT(resizedFitImage),ImageUtils.convertToSWT(imageArea.getCurrentImage()));
			}
		});
		@SuppressWarnings("unused")
		MenuItem separator4 = new MenuItem(submenuView, SWT.SEPARATOR);
		showHairMenu = new MenuItem(submenuView, SWT.CHECK);
		showHairMenu.setText(messages.getString("PROJECT_WINDOW_MENU_VIEW_HIDE-HAIR"));
		showHairMenu.addSelectionListener(new org.eclipse.swt.events.SelectionAdapter() {
			public void widgetSelected(org.eclipse.swt.events.SelectionEvent e) {
				tabs.changeHairVisibility(showHairMenu.getSelection());
				imageArea.redrawDrawingContainer();
			}
		});

		submenuEdit = new Menu(pushEdit);

		pushUndo = new MenuItem(submenuEdit, SWT.PUSH);
		pushRedo = new MenuItem(submenuEdit, SWT.PUSH);
		@SuppressWarnings("unused")
		MenuItem separator2 = new MenuItem(submenuEdit, SWT.SEPARATOR);
		pushClear = new MenuItem(submenuEdit, SWT.PUSH);

		pushUndo.setText(messages.getString("PROJECT_WINDOW_MENU_EDIT_UNDO") + "\tCtrl+Z");
		pushUndo.setAccelerator(SWT.CTRL + 'Z');
		pushUndo.setEnabled(false);
		pushUndo.setImage(new Image(Display.getCurrent(), StockIcon.getIcon(StockIcon.ICON_UNDO, StockIcon.SIZE_MENU, true)));
		pushUndo.addSelectionListener(new org.eclipse.swt.events.SelectionAdapter() {
			public void widgetSelected(org.eclipse.swt.events.SelectionEvent e) {
				tabs.undo();
				imageArea.redrawDrawingContainer();
			}
		});

		pushRedo.setText(messages.getString("PROJECT_WINDOW_MENU_EDIT_REDO") + "\tCtrl+Y");
		pushRedo.setAccelerator(SWT.CTRL + 'Y');
		pushRedo.setEnabled(false);
		pushRedo.setImage(new Image(Display.getCurrent(), StockIcon.getIcon(StockIcon.ICON_REDO, StockIcon.SIZE_MENU, true)));
		pushRedo.addSelectionListener(new org.eclipse.swt.events.SelectionAdapter() {
			public void widgetSelected(org.eclipse.swt.events.SelectionEvent e) {
				tabs.redo();
				imageArea.redrawDrawingContainer();
			}
		});

		pushClear.setText(messages.getString("PROJECT_WINDOW_MENU_EDIT_CLEAR") + "\tCtrl+K");
		pushClear.setAccelerator(SWT.CTRL + 'K');
		pushClear.setImage(new Image(Display.getCurrent(), StockIcon.getIcon(StockIcon.ICON_CLEAR, StockIcon.SIZE_MENU, true)));
		pushClear.addSelectionListener(new org.eclipse.swt.events.SelectionAdapter() {
			public void widgetSelected(org.eclipse.swt.events.SelectionEvent e) {
				tabs.clear();
				imageArea.redrawDrawingContainer();
			}
		});
		pushEdit.setMenu(submenuEdit);
		submenuProject = new Menu(pushProject);
		MenuItem push = new MenuItem(submenuProject, SWT.PUSH);
		push.setText(messages.getString("PROJECT_WINDOW_MENU_PROJECT_SAVE") + "\tCtrl+S");
		push.setAccelerator(SWT.CTRL + 'S');
		push.setImage(new Image(Display.getCurrent(), StockIcon.getIcon(StockIcon.ICON_SAVE, StockIcon.SIZE_MENU, true)));
		push.addSelectionListener(new org.eclipse.swt.events.SelectionAdapter() {
			public void widgetSelected(org.eclipse.swt.events.SelectionEvent e) {
				if (selectedFileForSave == null) { //does the same as Save as...
					String temporarySelectedFile = getFileToSave(messages.getString("PROJECT_WINDOW_SAVE_DIALOG_TITLE"));
					if (temporarySelectedFile != null) {
						saveToFile(temporarySelectedFile);
					}
				} else { //saves directly without asking
					saveToFile(selectedFileForSave);
				}
			}
		});
		MenuItem push1 = new MenuItem(submenuProject, SWT.PUSH);
		push1.setText(messages.getString("PROJECT_WINDOW_MENU_PROJECT_SAVEAS"));
		push1.setImage(new Image(Display.getCurrent(), StockIcon.getIcon(StockIcon.ICON_SAVE_AS, StockIcon.SIZE_MENU, true)));
		push1.addSelectionListener(new org.eclipse.swt.events.SelectionAdapter() {
			public void widgetSelected(org.eclipse.swt.events.SelectionEvent e) {
				String temporarySelectedFile = getFileToSave(messages.getString("PROJECT_WINDOW_SAVEAS_DIALOG_TITLE"));
				if (temporarySelectedFile != null) {
					saveToFile(temporarySelectedFile);
				}
			}
		});
		@SuppressWarnings("unused")
		MenuItem separator = new MenuItem(submenuProject, SWT.SEPARATOR);
		MenuItem push2 = new MenuItem(submenuProject, SWT.PUSH);
		push2.setText(messages.getString("PROJECT_WINDOW_MENU_PROJECT_PRINTPREVIEW"));
		push2.setImage(new Image(Display.getCurrent(), StockIcon.getIcon(StockIcon.ICON_PRINT_PREVIEW, StockIcon.SIZE_MENU, true)));
		push2.setEnabled(false);
		MenuItem push3 = new MenuItem(submenuProject, SWT.PUSH);
		push3.setText(messages.getString("PROJECT_WINDOW_MENU_PROJECT_PRINT") + "\tCtrl+P");
		push3.setAccelerator(SWT.CTRL + 'P');
		push3.setImage(new Image(Display.getCurrent(), StockIcon.getIcon(StockIcon.ICON_PRINT, StockIcon.SIZE_MENU, true)));
		push3.addSelectionListener(new org.eclipse.swt.events.SelectionAdapter() {
			public void widgetSelected(org.eclipse.swt.events.SelectionEvent e) {
				new dialogs.PrintDialog(sShell, resizedFitImage,imageArea.getCurrentImage());
			}
		});
		@SuppressWarnings("unused")
		MenuItem separator1 = new MenuItem(submenuProject, SWT.SEPARATOR);
		MenuItem push4 = new MenuItem(submenuProject, SWT.PUSH);
		push4.setText(messages.getString("PROJECT_WINDOW_MENU_PROJECT_CLOSE") + "\tCtrl+W");
		push4.setAccelerator(SWT.CTRL + 'W');

		ImageData closeImageData = StockIcon.getIcon(StockIcon.ICON_CLOSE, StockIcon.SIZE_MENU, true);
		Image closeMenuImage = null;
		if (closeImageData.width < 16) { //if close icon is smaller then create an transparent image with 16x16 and paint the icon on it
			org.eclipse.swt.graphics.Color black = sShell.getDisplay().getSystemColor(SWT.COLOR_BLACK);
			PaletteData palette = new PaletteData(new RGB[] { black.getRGB() });
			ImageData id = new ImageData(16, 16, 1, palette);
			id.transparentPixel = 0;
			closeMenuImage = new Image(Display.getCurrent(),id);
			GC g = new GC(closeMenuImage);
			g.drawImage(new Image(Display.getCurrent(),closeImageData), 2, 2);
			g.dispose();
		} else {
			closeMenuImage = new Image(Display.getCurrent(), closeImageData);
		}
		push4.setImage(closeMenuImage);
		push4.addSelectionListener(new org.eclipse.swt.events.SelectionAdapter() {
			public void widgetSelected(org.eclipse.swt.events.SelectionEvent e) {
				//TODO ask to save the project if necessary before closing
				sShell.close();
			}
		});
		pushProject.setMenu(submenuProject);
		sShell.setMenuBar(menuBar);
	}

	/**
	 * Asks the user to choose for a file for saving to using a FileDialog
	 * @param title the title of the dialog box
	 * @return the selected file or null if no file was selected
	 */
	private String getFileToSave(String title) {
		//create a file dialog to ask for the image to save
		FileDialog fd = new FileDialog(sShell, SWT.SAVE);
		fd.setText(title);
//		fd.setFilterNames(new String[] { FileDialogFilter.AllImagesName(), FileDialogFilter.jpegName(), FileDialogFilter.pngName() });
//		fd.setFilterExtensions(new String[] { FileDialogFilter.AllImagesExtension(), FileDialogFilter.jpegExtension(), FileDialogFilter.pngExtension() });
		String temporarySelectedFile = fd.open();
		if (temporarySelectedFile != null && new File(temporarySelectedFile).exists()) {
			MessageBox mb = new MessageBox(sShell, SWT.ICON_ERROR | SWT.OK | SWT.CANCEL);
			mb.setText(messages.getString("PROJECT_WINDOW_MESSAGE-BOX_OVERWRITE_TITLE"));
			//TODO MESSAGE1 should appear in bold
			mb.setMessage(messages.getString("PROJECT_WINDOW_MESSAGE-BOX_OVERWRITE_MESSAGE1") + " \"" + temporarySelectedFile + "\"" + "?\n\n" +
					messages.getString("PROJECT_WINDOW_MESSAGE-BOX_OVERWRITE_MESSAGE2"));
			int result = mb.open();
			if (result == SWT.CANCEL) {
				return null;
			}
		}
		return temporarySelectedFile;
	}

	/**
	 * Save the resulting image to the given file name
	 * @param fileName the file to save the image
	 */
	private void saveToFile(String fileName) {
		//add .png extension in the name as currently is the only format available for saving
		if (!fileName.endsWith(".png")) {
			fileName = fileName.concat(".png");
		}
		FileIOProgress fileIO = new FileIOProgress(sShell); 
		if (fileIO.saveImage(imageArea.getCurrentImage(), fileName)) {
			//TODO put * to project window's title when the image is different from the saved one
			sShell.setText(new File(fileName).getName() + " - " + messages.getString("PROJECT_WINDOW_TITLE"));
			selectedFileForSave = fileName;
		}
	}

	/**
	 * Returns the MainWindow's main container used as a parent for the tabs
	 * @return this project MainWindow's main container
	 */
	public Composite getParentMainContainer() {
		return parent.getMainContainer();
	}

	/**
	 * The shell used by this project
	 * @return the shell used by this project
	 */
	public Shell getSShell() {
		return sShell;
	}

	/**
	 * Enables the undo menu item if the argument is <code>true</code>,
	 * and disables it otherwise. A disabled menu item is typically
	 * not selectable from the user interface and draws with an
	 * inactive or "grayed" look.
	 *
	 * @param enabled the new enabled state
	 *
	 * @exception SWTException <ul>
	 *    <li>ERROR_WIDGET_DISPOSED - if the receiver has been disposed</li>
	 *    <li>ERROR_THREAD_INVALID_ACCESS - if not called from the thread that created the receiver</li>
	 * </ul>
	 */
	public void setEnabledUndo(boolean enabled) {
		pushUndo.setEnabled(enabled);
	}

	/**
	 * Enables the redo menu item if the argument is <code>true</code>,
	 * and disables it otherwise. A disabled menu item is typically
	 * not selectable from the user interface and draws with an
	 * inactive or "grayed" look.
	 *
	 * @param enabled the new enabled state
	 *
	 * @exception SWTException <ul>
	 *    <li>ERROR_WIDGET_DISPOSED - if the receiver has been disposed</li>
	 *    <li>ERROR_THREAD_INVALID_ACCESS - if not called from the thread that created the receiver</li>
	 * </ul>
	 */
	public void setEnabledRedo(boolean enabled) {
		pushRedo.setEnabled(enabled);
	}

	/**
	 * Notifies the undo menu item with the given event
	 * if the menu is enabled, otherwise nothing is done
	 * 
	 * @param e the event used to notify the menu item
	 * 
	 * @exception SWTException <ul>
	 *    <li>ERROR_WIDGET_DISPOSED - if the receiver has been disposed</li>
	 *    <li>ERROR_THREAD_INVALID_ACCESS - if not called from the thread that created the receiver</li>
	 * </ul>
	 */
	public void pushUndo(Event e) {
		if (pushUndo.isEnabled()) {
			pushUndo.notifyListeners(SWT.Selection, e);			
		}
	}

	/**
	 * Notifies the redo menu item with the given event
	 * if the menu is enabled, otherwise nothing is done
	 * 
	 * @param e the event used to notify the menu item
	 * 
	 * @exception SWTException <ul>
	 *    <li>ERROR_WIDGET_DISPOSED - if the receiver has been disposed</li>
	 *    <li>ERROR_THREAD_INVALID_ACCESS - if not called from the thread that created the receiver</li>
	 * </ul>
	 */
	public void pushRedo(Event e) {
		if (pushRedo.isEnabled()) {
			pushRedo.notifyListeners(SWT.Selection, e);
		}
	}

	/**
	 * Asks the image area to redraw it's current content
	 */
	public void redrawImageArea() {
		imageArea.redrawDrawingContainer();
	}

	/**
	 * Notifies the full screen menu item with the given event
	 * 
	 * @param e the event used to notify the menu item
	 * 
	 * @exception SWTException <ul>
	 *    <li>ERROR_WIDGET_DISPOSED - if the receiver has been disposed</li>
	 *    <li>ERROR_THREAD_INVALID_ACCESS - if not called from the thread that created the receiver</li>
	 * </ul>
	 */
	public void pushFullScreen(Event e) {
		pushFullScreen.notifyListeners(SWT.Selection, e);
	}

	/**
	 * Notifies the clear menu item with the given event
	 * 
	 * @param e the event used to notify the menu item
	 * 
	 * @exception SWTException <ul>
	 *    <li>ERROR_WIDGET_DISPOSED - if the receiver has been disposed</li>
	 *    <li>ERROR_THREAD_INVALID_ACCESS - if not called from the thread that created the receiver</li>
	 * </ul>
	 */
	public void pushClear(Event e) {
		pushClear.notifyListeners(SWT.Selection, e);
	}

	/**
	 * Enables the showHair menu item if the argument is <code>true</code>,
	 * and disables it otherwise. A disabled menu item is typically
	 * not selectable from the user interface and draws with an
	 * inactive or "grayed" look.
	 *
	 * @param enabled the new enabled state
	 *
	 * @exception SWTException <ul>
	 *    <li>ERROR_WIDGET_DISPOSED - if the receiver has been disposed</li>
	 *    <li>ERROR_THREAD_INVALID_ACCESS - if not called from the thread that created the receiver</li>
	 * </ul>
	 */
	public void setShowHairEnabled(boolean enabled) {
		showHairMenu.setEnabled(enabled);
	}

	/**
	 * Enables the eyes tracer menu item if the argument is <code>true</code>,
	 * and disables it otherwise. A disabled menu item is typically
	 * not selectable from the user interface and draws with an
	 * inactive or "grayed" look.
	 *
	 * @param enabled the new enabled state
	 *
	 * @exception SWTException <ul>
	 *    <li>ERROR_WIDGET_DISPOSED - if the receiver has been disposed</li>
	 *    <li>ERROR_THREAD_INVALID_ACCESS - if not called from the thread that created the receiver</li>
	 * </ul>
	 */
	public void setEyesTracerEnabled(boolean enabled) {
		pushEyesTracer.setEnabled(enabled);
	}

	/**
	 * Enables the lips tracer menu item if the argument is <code>true</code>,
	 * and disables it otherwise. A disabled menu item is typically
	 * not selectable from the user interface and draws with an
	 * inactive or "grayed" look.
	 *
	 * @param enabled the new enabled state
	 *
	 * @exception SWTException <ul>
	 *    <li>ERROR_WIDGET_DISPOSED - if the receiver has been disposed</li>
	 *    <li>ERROR_THREAD_INVALID_ACCESS - if not called from the thread that created the receiver</li>
	 * </ul>
	 */
	public void setLipsTracerEnabled(boolean enabled) {
		pushLipsTracer.setEnabled(enabled);
	}

	/**
	 * Enables the lens tracer menu item if the argument is <code>true</code>,
	 * and disables it otherwise. A disabled menu item is typically
	 * not selectable from the user interface and draws with an
	 * inactive or "grayed" look.
	 *
	 * @param enabled the new enabled state
	 *
	 * @exception SWTException <ul>
	 *    <li>ERROR_WIDGET_DISPOSED - if the receiver has been disposed</li>
	 *    <li>ERROR_THREAD_INVALID_ACCESS - if not called from the thread that created the receiver</li>
	 * </ul>
	 */
	public void setLensTracerEnabled(boolean enabled) {
		pushLensTracer.setEnabled(enabled);
	}

	/**
	 * Sets the selection state of the showHair menu item.
	 *
	 * @param selected the new selection state
	 *
	 * @exception SWTException <ul>
	 *    <li>ERROR_WIDGET_DISPOSED - if the receiver has been disposed</li>
	 *    <li>ERROR_THREAD_INVALID_ACCESS - if not called from the thread that created the receiver</li>
	 * </ul>
	 */
	public void setShowHairSelected(boolean selected) {
		//TODO the event should be sent even if the menu is disabled?
		Event e = new Event();
		e.widget = showHairMenu;
		showHairMenu.setSelection(selected);
		showHairMenu.notifyListeners(SWT.Selection, e);
	}

	/**
	 * Adds the specified mouse listener to receive mouse motion events
	 * from the project's drawing container. If listener is <code>null</code>,
	 * no exception is thrown and no action is performed.
	 * @param mouseListener the mouse listener
	 */
	public void registerMouseListener(MouseListener mouseListener) {
		imageArea.registerMouseListener(mouseListener);
	}

	/**
	 * Adds the specified mouse motion listener to receive mouse motion events
	 * from the project's drawing container. If listener is <code>null</code>,
	 * no exception is thrown and no action is performed.
	 * @param mouseMotionListener the mouse motion listener
	 */
	public void registerMouseMotionListener(MouseMotionListener mouseMotionListener) {
		imageArea.registerMouseMotionListener(mouseMotionListener);		
	}
}