/*******************************************************************************
 * 
 * jKiwi -- A virtual makeover and hairstyler application
 * Copyright (c) 2007-2008 Dan Mihai Ile, Maria Joao Roque Barbado Leal
 * 
 * This file is part of jKiwi.
 * 
 * jKiwi is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 * 
 * jKiwi is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 *******************************************************************************/

package tabs;

import java.awt.AlphaComposite;
import java.awt.Graphics2D;
import java.awt.Point;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;
import java.awt.geom.GeneralPath;
import java.awt.image.BufferedImage;
import java.util.ArrayList;
import java.util.ResourceBundle;


import org.eclipse.swt.SWT;
import org.eclipse.swt.events.DisposeEvent;
import org.eclipse.swt.events.DisposeListener;
import org.eclipse.swt.events.SelectionAdapter;
import org.eclipse.swt.events.SelectionEvent;
import org.eclipse.swt.graphics.Color;
import org.eclipse.swt.graphics.Image;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.widgets.Button;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Control;
import org.eclipse.swt.widgets.Display;
import org.eclipse.swt.widgets.Event;
import org.eclipse.swt.widgets.Label;
import org.eclipse.swt.widgets.Listener;
import org.eclipse.swt.widgets.Scale;

import core.Loader;

import utils.ImageUtils;
import utils.StockIcon;

/**
 * Instances of this class represents a tab that
 * can be used to paint the eyes colors, giving
 * the same effect of the contact lenses 
 *
 */
public class LensTab extends TabContainer {

	/**
	 * Instances of this class represents a state of the {@link LensTab}
	 */
	private class LensHistoryStep {

		/**
		 * The selected button
		 */
		private final ColorButton selectedColorButton;

		/**
		 * The opacity value used to paint the lens
		 */
		private final int lensOpacityScaleValue;

		/**
		 * Creates a new instance of this class to save the
		 * {@link LensTab} given state
		 * @param colorButton the selected button
		 * @param lensOpacityScaleValue the opacity value used to paint the lips
		 */
		public LensHistoryStep(ColorButton colorButton, int lensOpacityScaleValue) {

			//search for the reference of the actual button instead of the reference of the selectedButton
			ColorButton selectedColorButton = null;
			if (colorButton != null) {
				for (Control c : colorButton.getParent().getChildren()) {
					if (c instanceof ColorButton) {
						ColorButton cb = (ColorButton) c;
						if (cb.equals(colorButton)) {
							selectedColorButton = cb;
							break;
						}
					}
				}
			}
			this.selectedColorButton = selectedColorButton;
			this.lensOpacityScaleValue = lensOpacityScaleValue;
		}
	}

	/**
	 * The opacity that will be used to draw the lens color
	 */
	private static float ALPHA_COMPOSITE_OPACITY = 0.2f;

	/**
	 * Alpha composite used to draw the lens image
	 */
	private AlphaComposite currentCompositeOpacity;

	/**
	 * messages that are to be used by the entire GUI
	 */
	private ResourceBundle messages;

	/**
	 * A list with all the colors used by ColorButton
	 * For each color in this list one ColorButton will be created
	 */
	private Color[] colorsList = {
			new Color(Display.getCurrent(), 241, 209, 158),
			new Color(Display.getCurrent(), 231, 196, 141),
			new Color(Display.getCurrent(), 236, 185, 104),
			new Color(Display.getCurrent(), 234, 160, 44),
			new Color(Display.getCurrent(), 222, 145, 24),
			new Color(Display.getCurrent(), 200, 129, 19),
			new Color(Display.getCurrent(), 229, 201, 187),
			new Color(Display.getCurrent(), 203, 169, 152),
			new Color(Display.getCurrent(), 189, 137, 111),
			new Color(Display.getCurrent(), 166, 104, 73),
			new Color(Display.getCurrent(), 140, 79, 49),
			new Color(Display.getCurrent(), 114, 50, 18),
			new Color(Display.getCurrent(), 238, 205, 226),
			new Color(Display.getCurrent(), 217, 182, 214),
			new Color(Display.getCurrent(), 196, 144, 192),
			new Color(Display.getCurrent(), 193, 103, 186),
			new Color(Display.getCurrent(), 169, 70, 161),
			new Color(Display.getCurrent(), 129, 79, 104),
			new Color(Display.getCurrent(), 222, 229, 236),
			new Color(Display.getCurrent(), 206, 233, 240),
			new Color(Display.getCurrent(), 147, 219, 237),
			new Color(Display.getCurrent(), 80, 189, 220),
			new Color(Display.getCurrent(), 54, 154, 179),
			new Color(Display.getCurrent(), 23, 109, 130),
			new Color(Display.getCurrent(), 194, 241, 159),
			new Color(Display.getCurrent(), 160, 224, 115),
			new Color(Display.getCurrent(), 109, 181, 56),
			new Color(Display.getCurrent(), 36, 123, 31),
			new Color(Display.getCurrent(), 37, 102, 21),
			new Color(Display.getCurrent(), 37, 66, 14),
			new Color(Display.getCurrent(), 212, 212, 212),
			new Color(Display.getCurrent(), 184, 184, 184),
			new Color(Display.getCurrent(), 159, 157, 158),
			new Color(Display.getCurrent(), 114, 114, 114),
			new Color(Display.getCurrent(), 77, 77, 77),
			new Color(Display.getCurrent(), 0, 0, 0)
	};

	/**
	 * Used to set the upper and color buttons composite
	 */
	private TabGuiUtility tabGui;

	/**
	 * the tab's main composite
	 */
	private Composite mainComposite;

	/**
	 * All the colored buttons in use by this tab
	 */
	private ArrayList<ColorButton> coloredButtonList;

	//mouse listeners in the drawing area
	private MouseListener mouseListener;
	private MouseMotionListener mouseMotionListener;

	/**
	 * Composite for the bottom controls
	 */
	private Composite bottomComposite;

	//bottom composite label separators
	private Label labelSeparatorUp;
	private Label labelSeparatorDown;

	/**
	 * Scale for the opacity of the lips image
	 */
	private Scale opacityScale;

	/**
	 * Label for the opacity scale
	 */
	private Label labelOpacityScale;

	/**
	 * Button for clearing this tab's contents on the drawing area
	 */
	private Button clearButton;

	/**
	 * The eyes lens image created using the given eyes paths
	 */
	private BufferedImage eyesLensImage;

	/**
	 * flag to indicate whether the tab has or not access to call for redrawing the image area
	 */
	protected boolean imageAreaRedrawAllowed;

	/**
	 * A reference to this instance class
	 */
	private LensTab thisLensTab = this;

	/**
	 * Path that contains the eyes area
	 */
	private GeneralPath eyesGeneralPath;

	//store the center point of the left and right eye
	private Point leftEyeCenter;
	private Point rightEyeCenter;

	/**
	 * Constructs a new instance of this class given its parent,
	 * a style value describing its behavior and appearance and
	 * this instance's position in the {@link TabsContainer} array
	 *
	 * @param parent a {@link TabsContainer} which will be the parent of the new instance
	 * @param style the style of the tab
	 * @param tabLayerPosition the tab's position in the {@link TabsContainer} array
	 *
	 */
	public LensTab(TabsContainer parent, int style, int tabLayerPosition) {
		super(parent, style, tabLayerPosition);

		messages = Loader.getMessages();

		//the 2 things needed to be done by every implementation of TabComposite
		super.setTabImage(new Image(Display.getCurrent(), System.getProperty("program.pixmapspath") + "tab_icon_lens.png"));
		super.setTabToolTipText(messages.getString("LENS_TAB_TOOLTIP"));

		addDisposeListener(new DisposeListener() {
			public void widgetDisposed(DisposeEvent e) {
				for (Color c : colorsList) {
					c.dispose();
				}
			}
		});

		imageAreaRedrawAllowed = true;

		//create the tab's GUI
		tabGui = new TabGuiUtility(this);
		mainComposite = this;
		tabGui.createUpperComposite(super.getTabToolTipText());
		coloredButtonList = tabGui.createColorButtonsComposite(colorsList);
		tabGui.setSelectedColorButton(null);
		createBottomComposite();

		//default opacity
		currentCompositeOpacity = AlphaComposite.getInstance(AlphaComposite.SRC_OVER, ALPHA_COMPOSITE_OPACITY);

		extendColorButtonsBehavior();
	}

	/**
	 * This method initializes bottomComposite	
	 */
	public void createBottomComposite() {

		//grid data for the clear button
		GridData gridDataClearButton = new GridData();
		gridDataClearButton.grabExcessHorizontalSpace = true;
		gridDataClearButton.horizontalSpan = 2;
		gridDataClearButton.horizontalAlignment = org.eclipse.swt.layout.GridData.FILL;

		//grid data for the separator label up
		GridData gridDataSeparatorLabelUp = new GridData();
		gridDataSeparatorLabelUp.grabExcessHorizontalSpace = true;
		gridDataSeparatorLabelUp.horizontalSpan = 2;
		gridDataSeparatorLabelUp.horizontalAlignment = org.eclipse.swt.layout.GridData.FILL;

		//grid data for the separator label down
		GridData gridDataSeparatorLabelDown = new GridData();
		gridDataSeparatorLabelDown.horizontalAlignment = org.eclipse.swt.layout.GridData.FILL;
		gridDataSeparatorLabelDown.horizontalSpan = 2;
		gridDataSeparatorLabelDown.grabExcessHorizontalSpace = true;

		//grid data for the opacity scale
		GridData gridDataOpacityScale = new GridData();
		gridDataOpacityScale.grabExcessHorizontalSpace = true;
		gridDataOpacityScale.horizontalAlignment = org.eclipse.swt.layout.GridData.FILL;

		//grid data for the composite
		GridData gridDataComposite = new GridData();
		gridDataComposite.verticalAlignment = org.eclipse.swt.layout.GridData.END;
		gridDataComposite.grabExcessHorizontalSpace = true;
		gridDataComposite.horizontalAlignment = org.eclipse.swt.layout.GridData.FILL;
		gridDataComposite.grabExcessVerticalSpace = true;

		//grid layout for the composite
		GridLayout gridLayoutComposite = new GridLayout();
		gridLayoutComposite.numColumns = 2;

		//initialize the composite
		bottomComposite = new Composite(mainComposite, SWT.NONE);
		bottomComposite.setLayout(gridLayoutComposite);
		bottomComposite.setLayoutData(gridDataComposite);

		//initialize the upper label separator
		labelSeparatorUp = new Label(bottomComposite, SWT.SEPARATOR | SWT.HORIZONTAL);
		labelSeparatorUp.setLayoutData(gridDataSeparatorLabelUp);

		//initialize the clear button
		clearButton = new Button(bottomComposite, SWT.NONE);
		clearButton.setText(messages.getString("LENS_TAB_CLEAR"));
		clearButton.setImage(new Image(Display.getCurrent(), StockIcon.getIcon(StockIcon.ICON_CLEAR, StockIcon.SIZE_MENU, true)));
		clearButton.setLayoutData(gridDataClearButton);

		//initialize the opacity scale label		
		labelOpacityScale = new Label(bottomComposite, SWT.NONE);
		labelOpacityScale.setText(messages.getString("LENS_TAB_OPACITY"));

		//initialize the opacity scale
		opacityScale = new Scale(bottomComposite, SWT.NONE);
		opacityScale.setSelection(50);
		opacityScale.setLayoutData(gridDataOpacityScale);

		//initialize the lower separator label
		labelSeparatorDown = new Label(bottomComposite, SWT.SEPARATOR | SWT.HORIZONTAL);
		labelSeparatorDown.setLayoutData(gridDataSeparatorLabelDown);

		//clear button listener
		clearButton.addSelectionListener(new SelectionAdapter() {
			public void widgetSelected(SelectionEvent e) {
				//TODO check if history step is created only when needed
				if (e.widget instanceof Button && eyesLensImage != null) {
					tabsContainer.createHistoryStep();
					clearImageArea();
					if (imageAreaRedrawAllowed) {
						tabsContainer.redrawImageArea(thisLensTab);
					}
				}
			}
		});

		//opacity scale listeners
		opacityScale.addSelectionListener(new org.eclipse.swt.events.SelectionAdapter() {
			public void widgetSelected(org.eclipse.swt.events.SelectionEvent e) {
				double opacityInterval = 0.3;
				double adjustInterval = (((float)opacityScale.getSelection() / opacityScale.getMaximum()) * opacityInterval) - (opacityInterval / 2.0);
				currentCompositeOpacity = AlphaComposite.getInstance(AlphaComposite.SRC_OVER, (float)(ALPHA_COMPOSITE_OPACITY + adjustInterval));
				if (imageAreaRedrawAllowed) {
					tabsContainer.redrawImageArea(thisLensTab);
				}
			}
		});
		opacityScale.addListener(SWT.MouseDown, new Listener() {
			public void handleEvent(Event event) {
				tabsContainer.createHistoryStep();
			}
		});
	}

	/**
	 * Color buttons should do more than just updating the {@link ColorDifference}
	 * All color buttons listeners are implemented here
	 */
	private void extendColorButtonsBehavior() {
		for (final ColorButton button : coloredButtonList) {
			button.addMouseListener(new org.eclipse.swt.events.MouseAdapter() {

				public void mouseDown(org.eclipse.swt.events.MouseEvent e) {
					//TODO history is created but user may click cancel on EyesTracer and nothing is drawn
					tabsContainer.createHistoryStep();
				}

				public void mouseUp(org.eclipse.swt.events.MouseEvent e) {
					if (!hasEyesPath()) {
						tabsContainer.setLensPath();
					}
					createEyeLensImage();
					if (imageAreaRedrawAllowed) {
						tabsContainer.redrawImageArea(thisLensTab);
					}
				}
			});
		}
	}

	/**
	 * Clears this tab's image in the drawing area
	 * and resets the needed variables
	 */
	private void clearImageArea() {
		eyesLensImage = null;
		tabGui.setSelectedColorButton(null);
	}

	/**
	 * Create this tab's lens image  by
	 * drawing the lens path into a new
	 * image using the selected button's color
	 * and leave the center transparent
	 */
	private void createEyeLensImage() {
		eyesLensImage = null;
		if (eyesGeneralPath != null && tabGui.getSelectedButton() != null) {
			eyesLensImage = new BufferedImage(tabsContainer.getOriginalImage().getWidth(), tabsContainer.getOriginalImage().getHeight(), BufferedImage.TYPE_INT_RGB);
			Graphics2D g2d = eyesLensImage.createGraphics();
			g2d.setColor(tabGui.getCurrentColor());
			g2d.fillRect(0, 0, eyesLensImage.getWidth(), eyesLensImage.getHeight());
			g2d.dispose();
			eyesLensImage = ImageUtils.createSoftClipping(eyesLensImage, eyesGeneralPath);

			//clear the center
			g2d = eyesLensImage.createGraphics();
			g2d.setComposite(AlphaComposite.Clear);
			g2d.fillOval(leftEyeCenter.x - 3, leftEyeCenter.y - 3, 6, 6);
			g2d.fillOval(rightEyeCenter.x - 3, rightEyeCenter.y - 3, 6, 6);
			g2d.dispose();
		}
	}

	/**
	 * Set the lens path to the given one; if the new
	 * lens path is <code>null</code> nothing is done
	 * @param eyesPaths the new lens path or <code>null</code>
	 */
	public void setEyesPath(GeneralPath[] eyesPaths) {
		if (eyesPaths == null || eyesPaths[0] == null || eyesPaths[1] == null) return ;

		//recalculate the left and right center of the eyes path
		leftEyeCenter = new Point((int)eyesPaths[0].getBounds().getCenterX(), (int)eyesPaths[0].getBounds().getCenterY());
		rightEyeCenter = new Point((int)eyesPaths[1].getBounds().getCenterX(), (int)eyesPaths[1].getBounds().getCenterY());

		//create a general path for clipping
		eyesGeneralPath = new GeneralPath(eyesPaths[0]);
		eyesGeneralPath.append(eyesPaths[1], false);

		createEyeLensImage();

		if (imageAreaRedrawAllowed) {
			tabsContainer.redrawImageArea(thisLensTab);
		}
	}

	/**
	 * Check whether this tab instance has an associated eyes lens path
	 * @return true if the eyes lens path is found, false otherwise
	 */
	public boolean hasEyesPath() {
		return eyesGeneralPath != null;
	}

	@Override
	public BufferedImage getLayerImage() {
		//create a copy of this tab's image and return it
		BufferedImage transparentLensImage = null;
		if (eyesLensImage != null) {
			transparentLensImage = new BufferedImage(tabsContainer.getOriginalImage().getWidth(), tabsContainer.getOriginalImage().getHeight(), BufferedImage.TYPE_INT_ARGB);
			Graphics2D g2d = transparentLensImage.createGraphics();
			drawLensImage(g2d);
			g2d.dispose();
		}
		return transparentLensImage;
	}

	/**
	 * Draws the lens image using the given graphics context
	 * @param g2d the graphics context used for drawing operations
	 */
	private void drawLensImage(Graphics2D g2d) {

		//save the current composite
		java.awt.Composite defaultComposite = g2d.getComposite();

		//set lens image opacity
		g2d.setComposite(currentCompositeOpacity);

		//draw the lens image
		g2d.drawImage(eyesLensImage, 0, 0, null);

		//restore the g2d original composite
		g2d.setComposite(defaultComposite);

	}

	@Override
	public void drawCurrentLayer(Graphics2D g2d) {
		drawLensImage(g2d);
	}

	@Override
	public void clear() {
		clearImageArea();
	}

	@Override
	public Object getCurrentState() {
		return new LensHistoryStep(tabGui.getSelectedButton(),opacityScale.getSelection());
	}

	@Override
	public void setCurrentState(Object o) {
		if (o instanceof LensHistoryStep) {
			LensHistoryStep hs = (LensHistoryStep) o;

			imageAreaRedrawAllowed = false;

			tabGui.setSelectedColorButton(hs.selectedColorButton);

			opacityScale.setSelection(hs.lensOpacityScaleValue);
//			opacityScale.notifyListeners(eventType, event) //TODO this is how should work instead of the 3 lines below 
			double opacityInterval = 0.3;
			double adjustInterval = (((float)opacityScale.getSelection() / opacityScale.getMaximum()) * opacityInterval) - (opacityInterval / 2.0);
			currentCompositeOpacity = AlphaComposite.getInstance(AlphaComposite.SRC_OVER, (float)(ALPHA_COMPOSITE_OPACITY + adjustInterval));

			createEyeLensImage();

			imageAreaRedrawAllowed = true;
		}
	}


	public MouseListener mouseListener() {
		if (mouseListener == null) {
			mouseListener = new MouseListener() {

				//unimplemented methods
				public void mouseClicked(MouseEvent e) {}
				public void mouseExited(MouseEvent e) {}
				public void mousePressed(MouseEvent e) {}
				public void mouseReleased(MouseEvent e) {}

				public void mouseEntered(MouseEvent e) {
					e.getComponent().setCursor(new java.awt.Cursor(java.awt.Cursor.DEFAULT_CURSOR));					
				}
			};
		}
		return mouseListener;
	}

	public MouseMotionListener mouseMotionListener() {
		if (mouseMotionListener == null) {
			mouseMotionListener = new MouseMotionListener() {
				//unimplemented methods
				public void mouseDragged(MouseEvent e) {}
				public void mouseMoved(MouseEvent e) {}
			};
		}
		return mouseMotionListener;
	}

	@Override
	public void originalImageChanged() {
		eyesGeneralPath = null;		
	}

	public void disposeData() {
		eyesLensImage = null;
	}
}